@@ -1,4 +1,5 @@
 from ._private.address import get_bus_address, parse_address
+from ._private.util import replace_fds_with_idx, replace_idx_with_fds
 from .message import Message
 from .constants import BusType, MessageFlag, MessageType, ErrorType, NameFlag, RequestNameReply, ReleaseNameReply
 from .service import ServiceInterface
@@ -10,10 +11,10 @@ from . import introspection as intr
 from contextlib import suppress
 
 import inspect
-import traceback
 import socket
 import logging
 import xml.etree.ElementTree as ET
+import traceback
 
 from typing import Type, Callable, Optional, Union
 
@@ -51,9 +52,11 @@ class BaseMessageBus:
         self.unique_name = None
         self._disconnected = False
 
+        # True if the user disconnected himself, so don't throw errors out of
+        # the main loop.
+        self._user_disconnect = False
+
         self._method_return_handlers = {}
-        # buffer messages until connect
-        self._buffered_messages = []
         self._serial = 0
         self._user_message_handlers = []
         # the key is the name and the value is the unique name of the owner.
@@ -365,6 +368,7 @@ class BaseMessageBus:
 
         All pending  and future calls will error with a connection error.
         """
+        self._user_disconnect = True
         self._sock.shutdown(socket.SHUT_RDWR)
 
     def next_serial(self) -> int:
@@ -419,9 +423,6 @@ class BaseMessageBus:
 
         :param msg: The message to send.
         :type msg: :class:`Message <dbus_next.Message>`
-
-        :raises:
-            - :class:`Exception` - If a connection error occurred.
         """
         raise NotImplementedError('the "send" method must be implemented in the inheriting class')
 
@@ -451,7 +452,13 @@ class BaseMessageBus:
 
         return False
 
-    def _interface_signal_notify(self, interface, interface_name, member, signature, body):
+    def _interface_signal_notify(self,
+                                 interface,
+                                 interface_name,
+                                 member,
+                                 signature,
+                                 body,
+                                 unix_fds=[]):
         path = None
         for p, ifaces in self._path_exports.items():
             for i in ifaces:
@@ -466,7 +473,8 @@ class BaseMessageBus:
                                interface=interface_name,
                                member=member,
                                signature=signature,
-                               body=body))
+                               body=body,
+                               unix_fds=unix_fds))
 
     def _introspect_export_path(self, path):
         assert_object_path_valid(path)
@@ -504,8 +512,7 @@ class BaseMessageBus:
             ip_port = 0
 
             if transport == 'unix':
-                self._sock = socket.socket(socket.AF_UNIX,
-                                           socket.SOCK_STREAM | socket.SOCK_NONBLOCK)
+                self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                 self._stream = self._sock.makefile('rwb')
                 self._fd = self._sock.fileno()
 
@@ -518,6 +525,7 @@ class BaseMessageBus:
 
                 try:
                     self._sock.connect(filename)
+                    self._sock.setblocking(False)
                     break
                 except Exception as e:
                     err = e
@@ -693,9 +701,11 @@ class BaseMessageBus:
     @classmethod
     def _make_method_handler(cls, interface, method):
         def handler(msg, send_reply):
-            result = method.fn(interface, *msg.body)
-            body = ServiceInterface._fn_result_to_body(result, method.out_signature_tree)
-            send_reply(Message.new_method_return(msg, method.out_signature, body))
+            args = ServiceInterface._msg_body_to_args(msg)
+            result = method.fn(interface, *args)
+            body, fds = ServiceInterface._fn_result_to_body(
+                result, signature_tree=method.out_signature_tree)
+            send_reply(Message.new_method_return(msg, method.out_signature, body, fds))
 
         return handler
 
@@ -834,8 +844,13 @@ class BaseMessageBus:
                     raise DBusError(ErrorType.UNKNOWN_PROPERTY,
                                     'the property does not have read access')
                 prop_value = getattr(interface, prop.prop_getter.__name__)
+
+                body, unix_fds = replace_fds_with_idx(prop.signature, [prop_value])
+
                 send_reply(
-                    Message.new_method_return(msg, 'v', [Variant(prop.signature, prop_value)]))
+                    Message.new_method_return(msg,
+                                              'v', [Variant(prop.signature, body[0])],
+                                              unix_fds=unix_fds))
             elif msg.member == 'Set':
                 if not prop.access.writable():
                     raise DBusError(ErrorType.PROPERTY_READ_ONLY, 'the property is readonly')
@@ -844,12 +859,14 @@ class BaseMessageBus:
                     raise DBusError(ErrorType.INVALID_SIGNATURE,
                                     f'wrong signature for property. expected "{prop.signature}"')
                 assert prop.prop_setter
-                setattr(interface, prop.prop_setter.__name__, value.value)
+                body = replace_idx_with_fds(value.signature, [value.value], msg.unix_fds)
+                setattr(interface, prop.prop_setter.__name__, body[0])
                 send_reply(Message.new_method_return(msg))
 
         elif msg.member == 'GetAll':
-            result = self._get_all_properties(interface)
-            send_reply(Message.new_method_return(msg, 'a{sv}', [result]))
+            body, unix_fds = replace_fds_with_idx('a{sv}', [self._get_all_properties(interface)])
+            send_reply(Message.new_method_return(msg, 'a{sv}', body, unix_fds=unix_fds))
+
         else:
             assert False
 
